import argparse
import math
import time
import sys
import numpy as np
import torch
from torch.autograd import gradcheck
import logging
logging.basicConfig(level=logging.INFO)

TIME_SCALES = {'s': 1, 'ms': 1000, 'µs': 1000000}

def get_inputs(dtype, device, options):
    kwargs = {'dtype': dtype,
          'device': device,
          'requires_grad': True}
    input1 = torch.randn(options.x1_size, options.state_size, **kwargs)
    input2 = torch.randn(options.x2_size, options.state_size, **kwargs)
    return input1, input2, options.norm, options.eps

def _benchmark(device, dtype, distances, options):
    variables = get_inputs(dtype, device, options)

    # Warm-up / compilation
    distances(*variables)

    forward_min = math.inf
    forward_time = 0
    backward_min = math.inf
    backward_time = 0

    for _ in range(options.runs):
        variables[0].grad = None
        variables[1].grad = None

        start = time.time()
        d = distances(*variables)
        elapsed = time.time() - start
        forward_min = min(forward_min, elapsed)
        forward_time += elapsed

        start = time.time()
        d.sum().backward()
        elapsed = time.time() - start
        backward_min = min(backward_min, elapsed)
        backward_time += elapsed

    scale = TIME_SCALES[options.scale]
    forward_min *= scale
    backward_min *= scale
    forward_average = forward_time / options.runs * scale
    backward_average = backward_time / options.runs * scale
    return forward_min, forward_average, backward_min, backward_average

def benchmark(device, dtype, distances, options):
    forward_min, forward_average, backward_min, backward_average = _benchmark(device, dtype, distances, options)
    print('Forward: {0:.3f}/{1:.3f} {4} | Backward {2:.3f}/{3:.3f} {4} (min/avg)'.format(
        forward_min, forward_average, backward_min, backward_average,
        options.scale))

def check_gradient(device, dtype, distances, options):
    variables = get_inputs(dtype, device, options)
    if gradcheck(distances, variables):
        print('Ok')


def check_equal(first, second, verbose):
    if verbose:
        print()
    for i, (x, y) in enumerate(zip(first, second)):
        x = x.cpu().detach().numpy()
        y = y.cpu().detach().numpy()
        if verbose:
            print("x = {}".format(x.flatten()))
            print("y = {}".format(y.flatten()))
            print('-' * 80)

        rtol = 1e-4 if x.dtype == np.float32 else 1e-6
        # print("====", x.dtype, isinstance(x.dtype, np.float32), np.float64, np.float32, type(x.dtype))
        print(f"Vérification avec une tolérance {rtol:.3e}" )
        np.testing.assert_allclose(x, y, rtol=rtol, err_msg="Index: {}".format(i))

def check_forward(baseline_fn, tested_fn, variables, verbose):
    """Teste si la fonction testée donne les mêmes valeurs (forward)"""
    baseline_values = baseline_fn(*variables)
    cpp_values = tested_fn(*variables)

    print('Forward: Baseline (Python) vs. C++/CUDA ... ', end='')
    check_equal(baseline_values, cpp_values, verbose)
    print('Ok')

def zero_grad(variables):
    for variable in variables:
        if isinstance(variable, torch.Tensor):
            variable.grad.zero_()

def get_grads(variables):
    return [v.grad.clone() for v in variables if isinstance(v, torch.Tensor)]


def check_backward(baseline_fn, tested_fn, variables, verbose):
    """Teste si la fonction testée donne les mêmes valeurs (backward)"""

    # Calcul avec la baseline
    value = baseline_fn(*variables)
    noise = torch.rand_like(value)
    (noise * value).mean().backward()
    grad_baseline = get_grads(variables)

    # Calcul avec la fonction testée
    zero_grad(variables)
    cpp_value = tested_fn(*variables)
    (noise * cpp_value).mean().backward()
    grad_cpp = get_grads(variables)

    print('Backward: Baseline (Python) vs. C++/CUDA... ', end='')
    check_equal(grad_baseline, grad_cpp, verbose)
    print('Ok')


def check(device, dtype, distances, options):
    variables = get_inputs(dtype, device, options)

    from distances.naive import distances as baseline_fn

    if 'forward' in options.direction:
        check_forward(baseline_fn, distances, variables, options.verbose)

    if 'backward' in options.direction:
        check_backward(baseline_fn, distances, variables, options.verbose)


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('-1', '--x1-size', type=int, default=10, help="Taille du batch (x)")
    parser.add_argument('-2', '--x2-size', type=int, default=11, help="Taille du batch (y)")
    parser.add_argument('-s', '--state-size', type=int, default=20, help="Dimension de chaque vecteur")
    parser.add_argument('-c', '--cuda', action='store_true', help='Mode cuda')
    parser.add_argument('-t', '--threads', type=int, default=32, help='Nombre de threads par dimension (CUDA)')
    parser.add_argument('-d', '--double', action='store_true', help='Utiliser des doubles (sinon float)')
    parser.add_argument('-p', '--norm', type=int, default=2, help='Degré de la norme')
    parser.add_argument('-e', '--eps', type=float, default=1e-6, help='Epsilon (pour éviter les 0)')
    parser.add_argument('--verbose', action='store_true')

    subparsers = parser.add_subparsers()
    parser_benchmark = subparsers.add_parser('benchmark')
    parser_benchmark.add_argument('-r', '--runs', type=int, default=100)
    parser_benchmark.add_argument('--scale', choices=['s', 'ms', 'µs'], default='µs')
    parser_benchmark.add_argument('mode', choices=['py', 'cpp', 'cuda'])
    parser_benchmark.set_defaults(func=benchmark)

    parser_gradcheck = subparsers.add_parser('gradcheck')
    parser_gradcheck.add_argument('mode', choices=['py', 'cpp', 'cuda'])
    parser_gradcheck.set_defaults(func=check_gradient)

    parser_check = subparsers.add_parser('check')
    parser_check.add_argument('mode', choices=['cpp', 'cuda'])
    parser_check.set_defaults(func=check)
    parser_check.add_argument('direction', choices=['forward', 'backward'], nargs='+')

    options = parser.parse_args()

    device = torch.device("cuda") if options.cuda or options.mode == "cuda" else torch.device("cpu")
    dtype = torch.float64 if options.double or options.func == check_gradient else torch.float32
    logging.info("Device: %s / dtype: %s", device, dtype)

    if not options.func:
        sys.exit(0)

    if options.mode == "py":
        from distances.naive import distances
    else:
        from distances import distances
        import distances.functions as F
        F.NTHREADS = options.threads

    options.func(device, dtype, distances, options)

    print("---- FINISHED ----")
